/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#ifndef solver_h
#define solver_h

#include "SimpSolver.h"
#include "utils.h"
#include "un_simp.h"

using namespace Glucose;

extern SimpSolver* solver;
extern vec<Lit>    assumptions;

/* ========================================================================== */
/*                             Solver functions                               */
/* ========================================================================== */
void init_solver(bool elim_ll, bool verbose);
void ands2clauses(int time, bool elim_ll);
lbool check_sat(bool verb = 1);
lbool check_sat_with_restart(bool verb);
void assume(int lit);
void ucl(int a);
void bcl(int a, int b);
void tcl(int a, int b, int c);
void cl_lhs2rhs(unsigned lhs, std::list<unsigned> &rhs);
void cl_rhs2lhs(unsigned lhs, std::list<unsigned> &rhs);
void bigAnd2Solver(unsigned lhs, prop_tr_un_fl *ptuf);
void propagate_solver();

void ands2clauses_coi(int bad_lit, int time, bool verbose);

int import_unit_from_solver(bool verbose);
int import_binary_from_solver(bool verbose);

std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>> pt2cl(prop_tr_un_fl *ptuf);


#endif
